
/* 
 * afilter.s			
 *	
 * Two pole filter, implemented as an 8x10 matrix multiply
 *
 */

/* 8 bytes of state (only 4 bytes actually used) */

#define	POLEF_STATE_SIZE 7	/* Actually one less for DMA */ 	
	
/* Semi-permanent scalar registers */

.name	dm_in,		$21	# DMEM Input - points to samples
.name   dm_out,		$20	# DMEM output
.name	count,		$19	# Number to process
.name	state_addr,	$18	
.name	coef_base,	$16
.name	scale,		$14
.name	coef_index,	$13

/* Semi-permanent vector registers */

.name	vconst,		$v31	# Contains some usefule constants
.name	idata0,	 	$v30	# Two vectors of 8 inputs
.name 	odata0,		$v28  	# Two vectors of 8 outputs
.name	odataf,		$v26	# intermediate fractional value for output
.name	col0,		$v25	# Matrix columns
.name	col1,		$v24
.name	col2,		$v23
.name	col3,		$v22
.name	col4,		$v21
.name	col5,		$v20
.name	col6,		$v19
.name	col7,		$v18
.name	col8,		$v17
.name	vtmp0,		$v16
.name	vtmp1,		$v15
.name	vtmp2,		$v14


  # Get address of input and output

case_A_POLEF:
  # important constants:
  #
	lqv	vconst[0], VCONST_OFFSET(zero)
				vxor	odata0, odata0, odata0	# Clear last output
	lhu	dm_in, RSP_PARAMETER_DMEMIN(parbase)
				vxor	col8, col8, col8
	lhu	dm_out, RSP_PARAMETER_DMEMOUT(parbase)
				vxor	col7, col7, col7
	lhu	count, RSP_PARAMETER_COUNT(parbase)
				vxor	col6, col6, col6
  # Now if count is zero I don't have to do anything
	beq	count, zero, polefdone	
	# delay slot

  # Get the scale value from the command
	andi	scale, aud0, 0xffff
	mtc2	scale, vconst[10]	# byte aligned transfer
	sll	scale, scale, 2
	mtc2	scale, vtmp0[0]		# byte aligned transfer
	
  # Resolve DRAM address for state

.name	mask,		$1
.name	seg_id,		$2
.name	seg_addr,	$3

	lui	mask, 0x00ff			# Load up the state address
				vxor	col5, col5, col5
	ori	mask, mask, 0xffff
				vxor	col4, col4, col4
	and	state_addr, aud1, mask
				vxor	col3, col3, col3
	srl	seg_id, aud1, 24		# must clear bits
				vxor	col2, col2, col2
	sll	seg_id, seg_id, 2		# leave mult. by 4 for offset
	lw	seg_addr, RSP_SEG_OFFSET(seg_id)
	add	state_addr, state_addr, seg_addr

.unname mask
.unname seg_id
.unname seg_addr

  # Load state if required

.name	tmp0,	$1
.name	tmp1,	$2
.name	ssize,	$3

	slv	odata0[0], 0(scrbase)		# Store cleared last output
	srl	tmp0, aud0, 16			# Check A_INIT flag
	andi	tmp0, tmp0, A_INIT
	bgtz	tmp0, polef_no_load
	nop
  # Load up the state
	addi	tmp0, scrbase, 0
	addi	tmp1, state_addr, 0
	jal	DMAread
	addi	ssize, zero, POLEF_STATE_SIZE

 # Wait here for state to load up

polefwait1:	
	mfc0	$5, DMA_BUSY
	bne	$5, zero, polefwait1
	nop
 # release the semaphore
	mtc0	$0, SP_RESERVED

.unname	tmp1
.unname	ssize

polef_no_load:

  # Initialize registers
  # Address for coefficients

	addi	coef_index, zero, RSP_ADPCMTABLE_OFFSET

  # Register used for final scaling
	addi	tmp0, zero, 0x4
	mtc2	tmp0, vtmp2[0]

  # Load up the matrix coefficients into registers.
  # rescale col1 according to input scale

	lqv	col1[0], 16(coef_index)
	vmudm	vtmp0, col1, vtmp0[0]

  # Load up state
	ldv	odata0[8], 0(scrbase)
	
	sqv	vtmp0[0], 16(coef_index)

	lqv	col0[0], 0(coef_index)
	addi	coef_index, coef_index,-2
	lrv	col2[0], 32(coef_index)
	addi	coef_index, coef_index,-2
	lrv	col3[0], 32(coef_index)

	addi	coef_index, coef_index,-2
	lrv	col4[0], 32(coef_index)
	addi	coef_index, coef_index,-2
	lrv	col5[0], 32(coef_index)
	addi	coef_index, coef_index,-2
	lrv	col6[0], 32(coef_index)
	addi	coef_index, coef_index,-2
	lrv	col7[0], 32(coef_index)
	addi	coef_index, coef_index,-2
	lrv	col8[0], 32(coef_index)

	ldv	idata0[0], 0(dm_in)
	ldv	idata0[8], 8(dm_in)
.unname	tmp0

polefloop:
					vmudh	vtmp0, col0, odata0[6]
	addi	dm_in, dm_in, 16
					vmadh	vtmp0, col1, odata0[7]
	addi	count, count, -16
					vmadh	vtmp0, col2, idata0[0]
					vmadh	vtmp0, col3, idata0[1]
					vmadh	vtmp0, col4, idata0[2]
					vmadh	vtmp0, col5, idata0[3]
					vmadh	odata0, col6, idata0[4]
					vmadh	vtmp0, col7, idata0[5]
					vmadh	vtmp0, col8, idata0[6]
					vmadh	vtmp0, idata0, vconst[5]
	ldv	idata0[0], 0(dm_in)
					vsaw	odataf, vtmp1, odata0[1]
	ldv	idata0[8], 8(dm_in)
					vsaw	odata0, vtmp1, odata0[0]
 # Rescale first output
					vmudn	vtmp0, odataf, vtmp2[0]
					vmadh	odata0, odata0, vtmp2[0]
	sdv	odata0[0], 0(dm_out)
	sdv	odata0[8], 8(dm_out)
	bgtz	count, polefloop
	addi	dm_out, dm_out, 16

.name	tmp0,	$1
.name	tmp1,	$2
.name	ssize,	$3

	addi	tmp0, dm_out, -8	# Back to before the last 2 coefficients
	addi	tmp1, state_addr, 0
	jal	DMAwrite
	addi	ssize, zero, POLEF_STATE_SIZE
	
polefwait2:	
	mfc0	$5, DMA_BUSY
	bne	$5, zero, polefwait2
	nop

polefdone:
	j	AudDone
	mtc0	$0, SP_RESERVED

.unname tmp0
.unname tmp1
.unname	ssize

.unname	dm_in
.unname dm_out
.unname	count
.unname	state_addr
.unname	coef_base
.unname	coef_index
.unname scale

.unname vconst
.unname	idata0
.unname odata0
.unname	odataf
.unname	col0
.unname	col1
.unname	col2
.unname	col3
.unname	col4
.unname	col5
.unname	col6
.unname	col7
.unname	col8
.unname	vtmp0
.unname	vtmp1
.unname	vtmp2
