/*---------------------------------------------------------------------*
	Copyright (C) 1997, Nintendo.
	
	File		gxvtx_ex.s
	Coded    by	Yoshitaka Yasumoto.	Oct 29, 1997.
	
	$Id: gxvtx_ex.s,v 1.1.1.1 2002/05/02 03:29:11 blythe Exp $
 *---------------------------------------------------------------------*/

	#---------------------------------------------------------------------
	#  caseEX_G_VTX
	#	頂点データをロードして, 座標変換を行なう.
	#	Fast3D/F3DEX とは異なり, 頂点キャッシュ領域へ直接 DMA 転送
	#	を行ない, その後変換する. Vtx 構造体データは 16 Bytes/Vtx 
	#	で頂点キャッシュは 40 Bytes/Vtx なので, 処理対象の頂点キャ
	#	ッシュ領域の最終位置へ DMA 転送し, 変換処理を行なう.
	#	
	#	Level 2 から GBI が変更になったので注意すること.
	#	+------------+------+-----+-----+------+------------+
	#	|   G_VTX    | 0000 |     n     | 0000 | (v0+n)*2:8 |
	#	+-----+------+------+-----+-----+------+------------+
	#	|     | Seg  |                Address               |
	#	+-----+------+------------+------------+------------+
	#
	#	この仕様から gfx0 の b31-24 と b7-0 は 0x00 ではない
	#	ことが確定しているので, これを利用して LIGHT/MULMP フラグ
	#	の書き換えを行なう.
	#---------------------------------------------------------------------
caseEX_G_VTX:
	#-------------------------------------------------------------
	#  DMA 転送開始
	#-------------------------------------------------------------
  AssignForDMAproc
  FixedAssign(num,     1)
	lhu	dmem_adrs, RSP_SUBMOD_POINTS(gfx0)
	jal	AdrsFixup	# dram_adrs の設定
	lhu	num,       RSP_DLINPUT_BOTTOM-7(dinp)
	sub	dmem_adrs, dmem_adrs, num
	jal	DMAread		# dmem_adrs は必ず正なので READ 動作
	addi	dma_len, num, -1	
	#-------------------------------------------------------------
	#  DMA 転送待ちを利用してパラメータを取得する
	#	gmode_h:   Geometry Mode フラグの上位 16 bit
	#	num:       処理頂点数x2
	#	dest:      変換値の出力ポインタ
	#	src:       Vtx 構造体へのデータポインタ
	#-------------------------------------------------------------
  FixedAssign(src,    14)	/* gxlight_ex.s と共用 */
  FixedAssign(dest,   15)
  FixedAssign(lstat,  13)	/* gxlight_ex.s と共用 */
  FixedAssign(gmode_h, 5)	/* gxlight_ex.s と共用 */
  FixedAssign(lighton, 6)	/* gxlight_ex.s と共用 */
	lhu	gmode_h, RSP_LSTAT_RENDER(zero)
	srl	num,  num,  3		# num=n*2 を代入
	sub	dest, gfx0, num		# v0*2 の計算
	lhu	dest, RSP_SUBMOD_POINTS(dest)
	_mov	(src, dmem_adrs)
  EndAssignForDMAproc
	#-------------------------------------------------------------
	#  Geometry Mode を調べ, Light が ON ならライトの処理ルーチン
	#  へ Jump する. このとき Light モジュールを Overlay する可能
	#  性がある.
	#	ライトの変換処理ルーチンで v8-v18,19-20,gfx1,
	#	sys0,sys1,return が破壊される.
	#  DMA 転送待ちを利用してパラメータを取得する
	#	mulmpcalc: MPMTX 計算が必要かのフラグ
	#	lstat:     LSTAT 領域の先頭ポインタ(XfmLight でも設定)
	#	fogon:     Fog 処理をするかどうか (0 or 8)
	#  XfmLight において lighton にライトデータの最終位置が代入
	#  される. (RSP_LSTAT_LIGHT_NUM + RSP_SAVE_L0OKATX)
	#-------------------------------------------------------------
  FixedAssign(fogon, 7)
  FixedAssign(mulmpcalc, 8)
	lbu	mulmpcalc, RSP_LSTAT_MULMP_FLAG(zero)
#ifndef	NOLIGHT
	andi	lighton, gmode_h, G_LIGHTING_H
	bne	lighton, zero, caseEX_XfmLight
#endif
	andi	fogon, gmode_h, G_FOG_H		/* Delay */
	#-------------------------------------------------------------
	#  MMtx と PMtx を掛け合わせる必要があるかを判定し,
	#  不要なら SKIP する. 
	#	フラグに 0 以外の値を代入し処理済みにする.
	#	gfx0 の最下位バイトは 0 でないのでこれで OK
	#-------------------------------------------------------------
caseEX_G_VTX_MulMP:
	bne	mulmpcalc, zero, caseEX_G_VTX_GetMtx
	sll	fogon, fogon, 3		/* 8 倍する */
	sb	gfx0, RSP_LSTAT_MULMP_FLAG(zero) # フラグを非 0 にする
  EndAssign(mulmpcalc, 8)
	#-------------------------------------------------------------
	#  MMtx と PMtx との乗算処理を行ない, MPMtx を計算する.
	#	sys0,sys1,$20 が破壊される
	#-------------------------------------------------------------
  FixedAssign(mpmtx, 19)
  FixedAssign(mmtx,  20)
  FixedAssign(pmtx,  21)
	_li	(pmtx,  RSP_GSTAT_PMTX)
	_li	(mmtx,  RSP_GSTAT_MMTX)
	jal	math_MatCat
	_li	(mpmtx, RSP_GSTAT_MPMTX)
  EndAssign(mpmtx, 19)
  EndAssign(mmtx,  20)
  EndAssign(pmtx,  21)
	#-------------------------------------------------------------
	#  MPMtx を取得する.
	#  ViewPort の設定値の取得
	#  DMA の終了待ち
	#  dest2 の初期化
	#-------------------------------------------------------------
caseEX_G_VTX_GetMtx:
  Assign(mtx0, v8)
  Assign(mtx1, v9)
  Assign(mtx2, v10)
  Assign(mtx3, v11)
  Assign(mtf0, v12)
  Assign(mtf1, v13)
  Assign(mtf2, v14)
  Assign(mtf3, v15)
  Assign(vpscale, v16)
  Assign(vptrans, v17)
  Assign(txscale, v18)
  Assign(vpfog,   v19)
  Assign(vpscaleR,v21)
  Assign(dest2, 8)
#define	pNorm	txscale[4]
#define	clipBox	txscale[5]
						lqv	mtx0[0], RSP_GSTAT_MPMTX+ 0(zero)
						lqv	mtx2[0], RSP_GSTAT_MPMTX+16(zero)
						lqv	mtf0[0], RSP_GSTAT_MPMTX+32(zero)
						lqv	mtf2[0], RSP_GSTAT_MPMTX+48(zero)
	vadd	mtx1, mtx0, _0x0000		ldv	mtx1[0], RSP_GSTAT_MPMTX+ 8(zero)
	vadd	mtx3, mtx2, _0x0000		ldv	mtx3[0], RSP_GSTAT_MPMTX+24(zero)
	vadd	mtf1, mtf0, _0x0000		ldv	mtf1[0], RSP_GSTAT_MPMTX+40(zero)
	vadd	mtf3, mtf2, _0x0000		ldv	mtf3[0], RSP_GSTAT_MPMTX+56(zero)
						ldv	mtx0[8], RSP_GSTAT_MPMTX+ 0(zero)
						ldv	mtx2[8], RSP_GSTAT_MPMTX+16(zero)
				jal	caseEX_G_VTX_GetTrans
						ldv	mtf0[8], RSP_GSTAT_MPMTX+32(zero)
				jal	DMAwait
						ldv	mtf2[8], RSP_GSTAT_MPMTX+48(zero)
	
	#-------------------------------------------------------------
	#  頂点データ (x,y,z) を vin12 へ代入する.
	#-------------------------------------------------------------
  FixedAssign(vin12i, v20)
	ldv	vin12i[0],  0(src)		vmov	vpscale[5], vpscaleR[1]
	ldv	vin12i[8], 16(src)
  EndAssign(vpscaleR,v21)
	/*S*/
	#-------------------------------------------------------------
	#  頂点変換ループ開始
	#-------------------------------------------------------------
caseEX_G_VTX_Loop:
  FixedAssign(lptr,    9)
  FixedAssign(ldir1,   v2)
  FixedAssign(normX,   v7)
  FixedAssign(st12,    v22)	/* gxlight_ex.s と共用 */
  FixedAssign(vout12f, v23)
  FixedAssign(vout12i, v24)
  FixedAssign(vscn12i, v25)	/* gxlight_ex.s と共用 */
  FixedAssign(vscn12f, v26)
  FixedAssign(vfog12,  v27)	/* gxlight_ex.s と共用 */
	#-------------------------------------------------------------
	#  MP 行列を頂点座標に乗ずる
	#  スケーリング前のテクスチャ座標値をロードする
	#  必要ならライト処理を行なう
	#-------------------------------------------------------------
#ifndef	NOLIGHT
	vmudn	vtmp,    mtf3, _0x0001			lw	sys0,    28(src) # Clr2 のコピー
	vmadh	vtmp,    mtx3, _0x0001			llv	st12[12],  8(src)
	vmadn	vtmp,    mtf0, vin12i[0h]		_mov	(lptr, lighton)
	vmadh	vtmp,    mtx0, vin12i[0h]		lpv	ldir1[0],  oRSP_LSTAT_LIGHT_DIR_M+0(lptr)
	vmadn	vtmp,    mtf1, vin12i[1h]		sw	sys0,     8(src)
	vmadh	vtmp,    mtx1, vin12i[1h]		lpv	normX[0], 8(src)
	vmadn	vout12f, mtf2, vin12i[2h]		bne	lighton, zero, caseEX_Lighting
	vmadh	vout12i, mtx2, vin12i[2h]		/* Delay */
#else
	vmudn	vtmp,    mtf3, _0x0001			lw	sys0, 28(src)	# Clr2 のコピー
	vmadh	vtmp,    mtx3, _0x0001			llv	st12[12],  8(src)
	vmadn	vtmp,    mtf0, vin12i[0h]
	vmadh	vtmp,    mtx0, vin12i[0h]
	vmadn	vtmp,    mtf1, vin12i[1h]		sw	sys0,  8(src)
	vmadh	vtmp,    mtx1, vin12i[1h]
	vmadn	vout12f, mtf2, vin12i[2h]
	vmadh	vout12i, mtx2, vin12i[2h]
#endif
	#-------------------------------------------------------------
	#  前ループの FOG パラメータの計算
	#	値域を	[-0x0080,+0x7fff] から [0x7f00,0x7fff]にする.
	#	同様の処理を lighting 処理内で行なう.	
	#	In	vscn12i[3|7]
	#	Out	vfog12[3L|7L](確定)	
	#-------------------------------------------------------------
	vge	vfog12, vscn12i, _0x7f00		llv	st12[4], 24(src)
  EndAssign(lptr,   9)
  EndAssign(ldir1, v2)
  EndAssign(normX, v7)
caseEX_G_VTX_Return:
  FixedAssign(vclip12f, v7)
  FixedAssign(vclip12i, v6)
  FixedAssign(invW12f,  v5)
  FixedAssign(invW12i,  v4)
  FixedAssign(cflag,    10)
  FixedAssign(vtmp12f,  v21)
  FixedAssign(vtmp12i,  v2) 
	#-----------------------------------------------------
	#  NoN の場合 Z 値が負になる可能性があるので 0 で
	#  クランプする. このため NoN でない場合と比べ
	#  1-2 サイクルのロスとなる.
	#-----------------------------------------------------
#ifndef	NO_CLAMP_Z
  FixedAssign(vscz12i,   v3)
	vge	vscz12i, vscn12i, _0x0000
#else
	/* V*1 */
#endif
							#------------------------------------------	
							#  w に PerspNorm(=n) の値を乗する.		
							#	これにより 1/w の値を求めるときの	
							#	Overflow を防止する.			
							#------------------------------------------
	#------------------------------------------
	#  前ループのスクリーン座標値の出力
	#	Fog = OFF なら Ys の位置に書き込む.
	#  num 値を減算する
	#------------------------------------------
	addi	num, num, -2*2				vmudl	vtmp,    vout12f, pNorm
	sub	sys0, dest2, fogon			vmadm	vtmp12i, vout12i, pNorm
	sbv	vfog12[15], 19-40+8(sys0) /*FOG2*/	vmadn	vtmp12f, vzero,   _0x0000
								#----------------------------------
								#  クリップ用パラメータ作成
								#----------------------------------
#ifndef	NO_CLAMP_Z
	sbv	vfog12[7],  19-80+8(sys0) /*FOG1*/		vmov	vscn12f[1], vscz12i[2]
	ssv	vscz12i[12], 28-40(dest2) /*Zs2*/		vmudn	vclip12f, vout12f, clipBox
	slv	vscn12i[8],  24-40(dest2) /*XY2*/		vmadh	vclip12i, vout12i, clipBox
  EndAssign(vscz12i, v3)
#else
	sbv	vfog12[7],  19-80+8(sys0) /*FOG1*/		vmudn	vclip12f, vout12f, clipBox
	sdv	vscn12i[8],  24-40(dest2) /*XYZs2*/		vmadh	vclip12i, vout12i, clipBox
#endif
							#------------------------------------------
							#  1/nw の値を Newton 法で求める (1)
							#	vrcph/l で求めるのは 1/2X の値
							#	(R=1/X)
							#	1/X = R*(2-R*X) = R/2*(4+R/2*X*(-4))
							#------------------------------------------
	sdv	vscn12i[0],  24-80(dest2) /*XYZs1*/	vrcph	vtmp[0],    vtmp12i[3]
	ssv	vscn12f[12], 30-40(dest2) /*Zsf2*/	vrcpl	invW12f[3], vtmp12f[3]
#ifndef	NO_CLAMP_Z
	slv	vscn12f[2],  28-80(dest2) /*Zsif1*/	vrcph	invW12i[3], vtmp12i[7]
#else
	ssv	vscn12f[4],  30-80(dest2) /*Zsf1*/	vrcph	invW12i[3], vtmp12i[7]
#endif
	#-------------------------------------------
	#  dest ポインタ加算/ dest2 の取得
	#	dest += 80, dest2 = dest とする.
	#	もし num == -2 なら出力位置を重ねる.
	#  カラー値の取得
	#-------------------------------------------
  FixedAssign(vcolor,   v3)
		ldv	vcolor[0],   8(src)		vrcpl	invW12f[7], vtmp12f[7]
	sra	sys0,  num,  31				vrcph	invW12i[7], _0x0000
								#----------------------------------	
								#  リジェクト判定			
								#	W と X,Y,Z,-X,-Y,-Z を比較	
								#----------------------------------
	andi	sys0,  sys0, 40					vch	vtmp, vout12i, vout12i[3h]
	addi	dest,  dest, 80					vcl	vtmp, vout12f, vout12f[3h]
							#------------------------------------------
							#  1/nw の値を Newton 法で求める (2)		
							#------------------------------------------
	sub	dest2, dest, sys0			vmudl	vtmp,   vtmp12f, invW12f
	#------------------------------------------
	#   リジェクト判定値取得
	#	[3|7] は同じ W1,W2 を比較. この場合
	#	$vcc の b15,b11 は常に 1 となる.
	#	符号拡張で cflag の b31-16 は 1
	#	また W1<0 なら b3=1, W2<0 なら b7=1
	#	となる. これを NoN 版で使用する.
	#------------------------------------------
	cfc2	cflag, $vcc				vmadm	vtmp,   vtmp12i, invW12f
	#------------------------------------------
	#  マトリクス演算結果出力 / カラー値の出力
	#  Z 値の W へのコピー
	#  頂点データ (x,y,z) を vin12 へ代入する.
	#------------------------------------------
	sdv	vout12f[8],  8-40(dest2) /*XYZWf2*/	vmadn	vtmp12f, vtmp12f, invW12i
		ldv	vin12i[0],  0+32(src)		vmadh	vtmp12i, vtmp12i, invW12i
								#----------------------------------
								#  w の値の正負判定 (1)			
								#	w<0  なら vfog12=0x7fff	
								#	w>=0 なら vfog12=0x0000	
								#----------------------------------
	sdv	vout12f[0],  8-80(dest) /*XYZWf1*/		vge	vtmp, vout12i, _0x0000
							#------------------------------------------		
							#  1/nw の値を Newton 法で求める (3)			
							#------------------------------------------		
		lsv	vout12f[14],12-40(dest2)	vmudh	vtmp,    vone,   _0x0004	
	sdv	vout12i[8],  0-40(dest2) /*XYZWi2*/	vmadn	vscn12f, vtmp12f, _0xfffc
		lsv	vout12f[6], 12-80(dest)		vmadh	vscn12i, vtmp12i, _0xfffc
								#----------------------------------
								#  w の値の正負判定 (2)			
								#	In	$vcc			
								#	Out	vtmp12i
								#----------------------------------
	sdv	vout12i[0], 0-80(dest) /*XYZWi1*/		vmrg	vtmp12i, vzero, _0x7fff
							#----------------------------------
							#  クリップ判定	(1)		
							#	n*W と X,Y,Z,-X,-Y,-Z を比較
							#----------------------------------
	ldv	vin12i[8], 16+32(src)			vch	vtmp,  vout12i, vclip12i[3h]
							#------------------------------------------
							#  1/nw の値を Newton 法で求める (4)
							#	もし w が負なら 1/nw に 0x7fff^2 を
							#	加算し, 1/nw=0x7fff とする
							#------------------------------------------
	slv	vcolor[0], 16-40(dest2)	/*Clr2*/	vmudl	vtmp,     vscn12f, invW12f
		lsv	vout12i[14], 4-40(dest2)	vmadm	vtmp,     vscn12i, invW12f
	slv	vcolor[4], 16-80(dest)	/*Clr1*/	vmadn	invW12f,  vscn12f, invW12i
		lsv	vout12i[6],  4-80(dest)		vmadh	invW12i,  vscn12i, invW12i
	sh	cflag, 38-40(dest2)	/*CC2*/		vmadh	vtmp12i,  vtmp12i, _0x7fff
							#----------------------------------
							#  クリップ判定	(2)	
							#	n*W と X,Y,Z,-X,-Y,-Z を比較
							#----------------------------------
	sll	sys0, cflag, 4				vcl	vtmp,  vout12f, vclip12f[3h]
							#-------------------------------------------------------------
							#  透視変換を行なう (x,y,z を w で割る)
							#	1/nW を (x,y,z) に掛ける
							#-------------------------------------------------------------
	cfc2	cflag, $vcc				vmudl	vtmp,    vout12f, invW12f[3h]
	#------------------------------------------
	#   invW の書き込み
	#------------------------------------------
	ssv	invW12f[14], 34-40(dest2) /*invWf2*/	vmadm	vtmp,    vout12i, invW12f[3h]
	addi	src, src, 32				vmadn	vscn12f, vout12f, vtmp12i[3h]
	sh	cflag, 36-40(dest2)	  /*CC2*/	vmadh	vscn12i, vout12i, vtmp12i[3h]	
							#-----------------------------------
							#  テクスチャ座標値のスケーリング
							#-----------------------------------
	sll	cflag, cflag, 4				vmudm	vcolor, st12, txscale
	sh	sys0,  38-80(dest)			/* V */
	#-------------------------------------------	#-------------------------------------------------------------
	#   CC の書き込み				#  PerspNorm を掛け, (x/w,y/w,z/w) を求める
	#   [-nW|0000|+nW|0000|-W|0000|+W|0000]		#
	#-------------------------------------------	#-------------------------------------------------------------
	sh	cflag, 36-80(dest)			vmudl	vtmp,    vscn12f, pNorm
	ssv	invW12f[6],  34-80(dest)  /*invWf1*/	vmadm	vscn12i, vscn12i, pNorm
	ssv	invW12i[14], 32-40(dest2) /*invWi2*/	vmadn	vscn12f, vzero,   _0x0000
	ssv	invW12i[6],  32-80(dest)  /*invWi1*/	/* V */
							#-------------------------------------------------------------
							#  スクリーン座標系に変換する
							#	fog 値の計算も同時に行なう. このときの fog 計算値は
							#	0x7f00 を加算している.
							#-------------------------------------------------------------
	slv	vcolor[4],  20-40(dest2)  /*ST2*/	vmudh	vtmp,    vptrans, _0x0001
	slv	vcolor[12], 20-80(dest)	  /*ST1*/	vmadh	vtmp,    vpfog,   _0x7f00
	/* Rsvd */					vmadn	vscn12f, vscn12f, vpscale
	/* Rsvd */					bgtz	num, caseEX_G_VTX_Loop
	/* Delay */					vmadh	vscn12i, vscn12i, vpscale
	/* V */						/* S */
	/* V */						/* S */
	/* V */						bltz	return, caseEX_ClipRet # return<0 なら Clip へ戻る
#ifndef	NO_CLAMP_Z
	vge	vcolor, vscn12i, _0x0000		slv	vscn12i[8],  24-40(dest2) # Xs,Ys2
	vge	vfog12, vscn12i, _0x7f00		slv	vscn12i[0],  24-80(dest)  # Xs,Ys1
							ssv	vscn12f[12], 30-40(dest2) # Zsf2
							ssv	vscn12f[4],  30-80(dest)  # Zsf1
							ssv	vcolor[12],  28-40(dest2) # Zs2
							beq	fogon, zero, GfxDone
							ssv	vcolor[4],   28-80(dest)  # Zs1
							sbv	vfog12[15],  19-40(dest2) # Fog2
							j	GfxDone
							sbv	vfog12[7],   19-80(dest)  # Fog1
#else
	vge	vfog12, vscn12i, _0x7f00		sdv	vscn12i[8],  24-40(dest2) # Xs,Ys,Zs2
							sdv	vscn12i[0],  24-80(dest)  # Xs,Ys,Zs1
							ssv	vscn12f[12], 30-40(dest2) # Zsf2
							beq	fogon, zero, GfxDone
							ssv	vscn12f[4],  30-80(dest)  # Zsf1
							sbv	vfog12[15],  19-40(dest2) # Fog2
							j	GfxDone
							sbv	vfog12[7],   19-80(dest)  # Fog1
#endif
  EndAssign(vtmp12f, v21)
	#-------------------------------------------------------------
	#  ViewPort の設定値の取得
	#  FogFactor  の値の取得 (vpscale[3|7],vptrans[3|7] へ代入)
	#  PerpsNorm  の値の取得 (txscale[4] へ代入)
	#  ClipBox    の値の取得 (txscale[5] へ代入)
	#  Fog 計算用パラメータ vpfog の取得
	#	[3|7] のみ 0x0001 となっている. その他は 0x0000
	#  Clipボックスの係数/Texture のスケール値の取得 [2|3|6|7] へ
	#-------------------------------------------------------------					
  Assign(vpscaleR, v21)
caseEX_G_VTX_GetTrans:				_li	(lstat, RSP_LSTAT_OFFSET)
						ldv	vpscale[0], RSP_GSTAT_VIEWPORT_SC(zero)
						ldv	vpscale[8], RSP_GSTAT_VIEWPORT_SC(zero)
						llv	vtmp[0],    oRSP_LSTAT_FOG_FACTOR(lstat)
						ldv	vptrans[0], RSP_GSTAT_VIEWPORT_TX(zero)
	/* Resv */				ldv	vptrans[8], RSP_GSTAT_VIEWPORT_TX(zero)
	vlt	vpfog,    vconst0, _0x7f00	/* Resv */
	vsub	vpscaleR, vzero,   vpscale	llv	txscale[4],  oRSP_LSTAT_TEX_SCALE(lstat)
	vmrg	vpscale,  vpscale, vtmp[0]	llv	txscale[12], oRSP_LSTAT_TEX_SCALE(lstat)
	vmrg	vpfog,    vzero,   _0x0001	llv	txscale[8],  RSP_GSTAT_VPERSPNORM(zero)	# lsv では不届
	vmrg	vptrans,  vptrans, vtmp[1]	lsv	txscale[10], oRSP_LSTAT_CLIPSELECT+6(lstat)
	vmov	vpscale[1], vpscaleR[1]		jr	return		# Vp.y の反転 
	/* Delay */				addi	dest2, outp, 80
  EndAssign(num,     1)
  EndAssign(src,    14)
  EndAssign(dest,   15)
  EndAssign(gmode_h, 5)
  EndAssign(lighton, 6)
  EndAssign(fogon,   7)
  EndAssign(dest2,   8)
  EndAssign(cflag,  10)
  EndAssign(lstat,  13)
  EndAssign(mtx0, v8)
  EndAssign(mtx1, v9)
  EndAssign(mtx2, v10)
  EndAssign(mtx3, v11)
  EndAssign(mtf0, v12)
  EndAssign(mtf1, v13)
  EndAssign(mtf2, v14)
  EndAssign(mtf3, v15)
  EndAssign(vpscale, v16)
  EndAssign(vptrans, v17)
  EndAssign(txscale, v18)
  EndAssign(vpfog,   v19)
  EndAssign(vin12i,  v20)
  EndAssign(vpscaleR,v21)
  EndAssign(st12,    v22)
  EndAssign(vout12f, v23)
  EndAssign(vout12i, v24)
  EndAssign(vscn12i, v25)
  EndAssign(vscn12f, v26)
  EndAssign(vfog12,  v27)
  EndAssign(vclip12f, v7)
  EndAssign(vclip12i, v6)
  EndAssign(invW12f,  v5)
  EndAssign(invW12i,  v4)
  EndAssign(vcolor,   v3)
  EndAssign(vtmp12i,  v2) 
	
/*======== End of gxvtx_ex.s ========*/
