/*---------------------------------------------------------------------*
	Copyright (C) 1998, Nintendo.
	
	File		gxvtx_lxrej.s
	Coded    by	Yoshitaka Yasumoto.	Feb 25, 1998.
	
	$Id: gxvtx_lxrej.s,v 1.1.1.1 2002/05/02 03:29:11 blythe Exp $
 *---------------------------------------------------------------------*/
 #!Reserved  (0,11,12,22,23,24,25,26,27,31)
 #!Reserved  (v0,v1,v29,v30,v31)
	
#define	NO_CLAMP_Z	/* No Clamp */ 
	
	#---------------------------------------------------------------------
	#  caseLXRej_G_VTX
	#	頂点データをロードして, 座標変換を行なう.
	#	Fast3D/F3DEX とは異なり, 頂点キャッシュ領域へ直接 DMA 転送
	#	を行ない, その後変換する. Vtx 構造体データは 16 Bytes/Vtx 
	#	で頂点キャッシュは 40 Bytes/Vtx なので, 処理対象の頂点キャ
	#	ッシュ領域の最終位置へ DMA 転送し, 変換処理を行なう.
	#	
	#	Level 2 から GBI が変更になったので注意すること.
	#	+------------+------+-----+-----+------+------------+
	#	|   G_VTX    | 0000 |     n     | 0000 | (v0+n)*2:8 |
	#	+-----+------+------+-----+-----+------+------------+
	#	|     | Seg  |                Address               |
	#	+-----+------+------------+------------+------------+
	#
	#	この仕様から gfx0 の b31-24 と b7-0 は 0x00 ではない
	#	ことが確定しているので, これを利用して LIGHT/MULMP フラグ
	#	の書き換えを行なう.
	#---------------------------------------------------------------------
caseLXRej_G_VTX:
	#-------------------------------------------------------------
	#  DMA 転送開始
	#-------------------------------------------------------------
  AssignForDMAproc
  FixedAssign(num,     1)
	lhu	dmem_adrs, RSP_SUBMOD_POINTS(gfx0)
	jal	AdrsFixup			# dram_adrs の設定
	lhu	num,       RSP_DLINPUT_BOTTOM-7(dinp)
	andi	dmem_adrs, dmem_adrs, 0xff8	# 8 バイト境界へ
	sub	dmem_adrs, dmem_adrs, num
	jal	DMAread		# dmem_adrs は必ず正なので READ 動作
	addi	dma_len, num, -1
	#-------------------------------------------------------------
	#  DMA 転送待ちを利用してパラメータを取得する
	#	gmode_h:   Geometry Mode フラグの上位 16 bit
	#	num:       処理頂点数x2
	#	dest:      変換値の出力ポインタ
	#	src:       Vtx 構造体へのデータポインタ
	#-------------------------------------------------------------
  FixedAssign(src,    14)	/* gxlight_lxrej.s と共用 */
  FixedAssign(dest,   15)
  FixedAssign(lstat,  13)	/* gxlight_lxrej.s と共用 */
  FixedAssign(gmode_h, 5)	/* gxlight_lxrej.s と共用 */
  FixedAssign(lighton, 6)	/* gxlight_lxrej.s と共用 */
	lhu	gmode_h, RSP_LSTAT_RENDER(zero)
	srl	num,  num,  3		# num=n*2 を代入
	sub	dest, gfx0, num		# v0*2 の計算
	lhu	dest, RSP_SUBMOD_POINTS(dest)
	_mov	(src, dmem_adrs)
  EndAssignForDMAproc
	#-------------------------------------------------------------
	#  Geometry Mode を調べ, Light が ON ならライトの処理ルーチン
	#  へ Jump する. このとき Light モジュールを Overlay する可能
	#  性がある.
	#	ライトの変換処理ルーチンで v8-v18,19-20,gfx1,
	#	sys0,sys1,return が破壊される.
	#  DMA 転送待ちを利用してパラメータを取得する
	#	mulmpcalc: MPMTX 計算が必要かのフラグ
	#	lstat:     LSTAT 領域の先頭ポインタ(XfmLight でも設定)
	#	fogon:     Fog 処理をするかどうか (0 or 8)
	#  XfmLight において lighton にライトデータの最終位置が代入
	#  される. (RSP_LSTAT_LIGHT_NUM + RSP_SAVE_L0OKATX)
	#-------------------------------------------------------------
  FixedAssign(fogon, 7)
  FixedAssign(mulmpcalc, 8)
	lbu	mulmpcalc, RSP_LSTAT_MULMP_FLAG(zero)
	andi	lighton, gmode_h, G_LIGHTING_H
	bne	lighton, zero, caseLXRej_XfmLight
	andi	fogon, gmode_h, G_FOG_H		/* Delay */
	#-------------------------------------------------------------
	#  MMtx と PMtx を掛け合わせる必要があるかを判定し,
	#  不要なら SKIP する. 
	#	フラグに 0 以外の値を代入し処理済みにする.
	#	gfx0 の最下位バイトは 0 でないのでこれで OK
	#-------------------------------------------------------------
caseLXRej_G_VTX_MulMP:
	bne	mulmpcalc, zero, caseLXRej_G_VTX_GetMtx
	sll	fogon, fogon, 3		/* 8 倍する */
	sb	gfx0, RSP_LSTAT_MULMP_FLAG(zero) # フラグを非 0 にする
  EndAssign(mulmpcalc, 8)
	#-------------------------------------------------------------
	#  MMtx と PMtx との乗算処理を行ない, MPMtx を計算する.
	#	sys0,sys1,$20 が破壊される
	#-------------------------------------------------------------
  FixedAssign(mpmtx, 19)
  FixedAssign(mmtx,  20)
  FixedAssign(pmtx,  21)
	_li	(pmtx,  RSP_GSTAT_PMTX)		
	_li	(mmtx,  RSP_GSTAT_MMTX)
	jal	math_MatCat
	_li	(mpmtx, RSP_GSTAT_MPMTX)
  EndAssign(mpmtx, 19)
  EndAssign(mmtx, 20)
  EndAssign(pmtx, 21)		
	#-------------------------------------------------------------
	#  MPMtx を取得する.
	#  ViewPort の設定値の取得
	#  FogFactor  の値の取得 (vpscale[3|7],vptrans[3|7] へ代入)
	#  PerpsNorm  の値の取得 (txscale[4] へ代入)
	#  ClipBox    の値の取得 (txscale[5] へ代入)
	#  Fog 計算用パラメータ vpfog の取得
	#	[3|7] のみ 0x0001 となっている. その他は 0x0000
	#  Clipボックスの係数/Texture のスケール値の取得 [2|3|6|7] へ
	#-------------------------------------------------------------					
caseLXRej_G_VTX_GetMtx:
  FixedAssign(mtx0, v8)
  FixedAssign(mtx1, v9)
  FixedAssign(mtx2, v10)
  FixedAssign(mtx3, v11)
  FixedAssign(mtf0, v12)
  FixedAssign(mtf1, v13)
  FixedAssign(mtf2, v14)
  FixedAssign(mtf3, v15)
  FixedAssign(vpscale, v16)
  FixedAssign(vptrans, v17)
  FixedAssign(txscale, v18)
  FixedAssign(vpfog,   v19)
  Assign(vpscaleR, v2)
  Assign(vptransY,v20)
  FixedAssign(dest2,  8)
  FixedAssign(dest2x, 10)
#define	pNorm	txscale[4]
#define	clipBox	txscale[5]
						_li	(lstat, RSP_LSTAT_OFFSET)
						ldv	vpscale[0], RSP_GSTAT_VIEWPORT_SC(zero)
						ldv	vpscale[8], RSP_GSTAT_VIEWPORT_SC(zero)
						llv	vtmp[0],    oRSP_LSTAT_FOG_FACTOR(lstat)
#ifdef	NOSUBPIX
	vadd	vptransY, vone, _0x0001		ldv	vptrans[0], RSP_GSTAT_VIEWPORT_TX(zero)	
	/* Resv */				ldv	vptrans[8], RSP_GSTAT_VIEWPORT_TX(zero)
	vlt	vpfog,    vconst0, _0x7f00	/* Resv */
	vsub	vpscaleR, vzero,   vpscale	llv	txscale[4],  oRSP_LSTAT_TEX_SCALE(lstat)
	vmrg	vpscale,  vpscale, vtmp[0]	llv	txscale[12], oRSP_LSTAT_TEX_SCALE(lstat)
	vmrg	vpfog,    vzero,   _0x0001	llv	txscale[8],  RSP_GSTAT_VPERSPNORM(zero)	# lsv では不届
	vadd	vptransY, vptrans, vptransY	lsv	txscale[10], oRSP_LSTAT_CLIPSELECT+6(lstat)
	vmrg	vptrans,  vptrans, vtmp[1]	addi	dest2, outp, 80
	vmov	vpscale[1], vpscaleR[1]		lqv	mtx0[0], RSP_GSTAT_MPMTX+ 0(zero)
	vmov	vpscale[5], vpscaleR[1]		lqv	mtx2[0], RSP_GSTAT_MPMTX+16(zero)
	vmov	vptrans[1], vptransY[1]		lqv	mtf0[0], RSP_GSTAT_MPMTX+32(zero)
	vmov	vptrans[5], vptransY[1]		lqv	mtf2[0], RSP_GSTAT_MPMTX+48(zero)
#else
						ldv	vptrans[0], RSP_GSTAT_VIEWPORT_TX(zero)	
	/* Resv */				ldv	vptrans[8], RSP_GSTAT_VIEWPORT_TX(zero)
	vlt	vpfog,    vconst0, _0x7f00	/* Resv */
	vsub	vpscaleR, vzero,   vpscale	llv	txscale[4],  oRSP_LSTAT_TEX_SCALE(lstat)
	vmrg	vpscale,  vpscale, vtmp[0]	llv	txscale[12], oRSP_LSTAT_TEX_SCALE(lstat)
	vmrg	vpfog,    vzero,   _0x0001	llv	txscale[8],  RSP_GSTAT_VPERSPNORM(zero)	# lsv では不届
						lsv	txscale[10], oRSP_LSTAT_CLIPSELECT+6(lstat)
	vmrg	vptrans,  vptrans, vtmp[1]	addi	dest2, outp, 80
	vmov	vpscale[1], vpscaleR[1]		lqv	mtx0[0], RSP_GSTAT_MPMTX+ 0(zero)
	vmov	vpscale[5], vpscaleR[1]		lqv	mtx2[0], RSP_GSTAT_MPMTX+16(zero)
						lqv	mtf0[0], RSP_GSTAT_MPMTX+32(zero)
						lqv	mtf2[0], RSP_GSTAT_MPMTX+48(zero)
#endif
	vadd	mtx1, mtx0, _0x0000		ldv	mtx1[0], RSP_GSTAT_MPMTX+ 8(zero)
	vadd	mtx3, mtx2, _0x0000		ldv	mtx3[0], RSP_GSTAT_MPMTX+24(zero)
	vadd	mtf1, mtf0, _0x0000		ldv	mtf1[0], RSP_GSTAT_MPMTX+40(zero)
	vadd	mtf3, mtf2, _0x0000		ldv	mtf3[0], RSP_GSTAT_MPMTX+56(zero)
						ldv	mtx0[8], RSP_GSTAT_MPMTX+ 0(zero)
						ldv	mtx2[8], RSP_GSTAT_MPMTX+16(zero)
						ldv	mtf0[8], RSP_GSTAT_MPMTX+32(zero)
						jal	DMAwait
						ldv	mtf2[8], RSP_GSTAT_MPMTX+48(zero)
  EndAssign(vpscaleR, v2)
  EndAssign(vptransY,v20)
	#-------------------------------------------------------------
	#  頂点データ (x,y,z) を vin12 へ代入する.
	#-------------------------------------------------------------
  FixedAssign(vin12i, v20)
	ldv	vin12i[0],  0(src)
	ldv	vin12i[8], 16(src)		
	addi	dest2x, dest2, 4
	#-------------------------------------------------------------
	#  頂点変換ループ開始
	#-------------------------------------------------------------
caseLXRej_G_VTX_Loop:
  FixedAssign(lptr,    9)
  FixedAssign(ldir1,   v2)
  FixedAssign(normX,   v7)
  FixedAssign(st12,    v22)	/* gxlight_ex.s と共用 */
  FixedAssign(vout12f, v23)
  FixedAssign(vout12i, v24)
  FixedAssign(vscn12i, v25)	/* gxlight_ex.s と共用 */
  FixedAssign(vscn12f, v26)
  FixedAssign(vfog12,  v27)	/* gxlight_ex.s と共用 */
	#-------------------------------------------------------------
	#  MP 行列を頂点座標に乗ずる
	#  スケーリング前のテクスチャ座標値をロードする
	#  必要ならライト処理を行なう
	#-------------------------------------------------------------
	vmudn	vtmp,    mtf3, _0x0001			lw	sys0,    28(src) # Clr2 のコピー
	vmadh	vtmp,    mtx3, _0x0001				llv	st12[12],  8(src)
	vmadn	vtmp,    mtf0, vin12i[0h]		_mov	(lptr, lighton)
#ifdef	ALPHA_LIGHTING
	vmadh	vtmp,    mtx0, vin12i[0h]			lqv	ldir1[0],  32(lptr)
#else
	vmadh	vtmp,    mtx0, vin12i[0h]			lpv	ldir1[0],  oRSP_LSTAT_LIGHT_DIR_M+0(lptr)
#endif
	vmadn	vtmp,    mtf1, vin12i[1h]		sw	sys0,     8(src)
	vmadh	vtmp,    mtx1, vin12i[1h]		lpv	normX[0], 8(src)
	vmadn	vout12f, mtf2, vin12i[2h]		bne	lighton, zero, caseLXRej_Lighting
	vmadh	vout12i, mtx2, vin12i[2h]		/* Delay */
	#-------------------------------------------------------------
	#  前ループの FOG フォグ値のクランプ
	#-------------------------------------------------------------
	vge	vfog12,  vscn12i, _0x7f00		llv	st12[4], 24(src)
  EndAssign(lptr, 9)
  EndAssign(ldir1, v2)
  EndAssign(normX, v7)

#ifdef	NOSUBPIX	/* LXRej */
	#-------------------------------------------------------------
	#  前ループの Y 座標のサブピクセルのマスク
	#-------------------------------------------------------------
caseLXRej_G_VTX_Return:
  Assign(yscn12i, v2)
	vand	yscn12i, vscn12i, _0xfffc		/*S*/
	#-------------------------------------------------------------
	#  w に PerspNorm(=n) の値を乗する.
	#	これにより 1/w の値を求めるときの	
	#	Overflow を防止する.
	#  前ループのスクリーン座標値の出力
	#-------------------------------------------------------------
  Assign(cflag, 2)
  Assign(W12f, v3)
  Assign(W12i, v4)
	vmudl	vtmp, vout12f, pNorm			addi	num, num, -2*2		# num 値を減算する
#ifndef	NO_CLAMP_Z
	vmadm	W12i, vout12i, pNorm			ssv	vscn12i[8], oRSP_POINT_XS-RSP_POINT_LEN*1(dest2)    /*Xs2*/
	vmadn	W12f, vzero,   _0x0000			ssv	vscn12i[0], oRSP_POINT_XS-RSP_POINT_LEN*2(dest2)    /*Xs1*/
#else
	vmadm	W12i, vout12i, pNorm			sdv	vscn12i[8], oRSP_POINT_XS-RSP_POINT_LEN*1-4(dest2x) /*XZs2*/
	vmadn	W12f, vzero,   _0x0000			sdv	vscn12i[0], oRSP_POINT_XS-RSP_POINT_LEN*2(dest2)    /*XZs1*/
#endif
	#----------------------------------------------
	#  クリップ用パラメータ作成	
	#----------------------------------------------
  Assign(vclip12f, v5)
  Assign(vclip12i, v6)
	vmudn	vclip12f, vout12f, clipBox		ssv	vscn12f[12],oRSP_POINT_ZSF-RSP_POINT_LEN*1(dest2) /*Zsf2*/
	vmadh	vclip12i, vout12i, clipBox		ssv	vscn12f[4], oRSP_POINT_ZSF-RSP_POINT_LEN*2(dest2) /*Zsf1*/
	#----------------------------------------------
	#  1/nw の値を Newton 法で求める (1)
	#	vrcph/l で求めるのは 1/2X の値
	#	(R=1/X)
	#	1/X = R*(2-R*X) = R/2*(4+R/2*X*(-4))
	#----------------------------------------------
  Assign(invW12f, v7)
  Assign(invW12i, v21)
	/*Resv*/					sub	sys0, dest2, fogon	# Fog=OFF なら Ys の位置に書き込む.
caseLXRej_G_VTX_Return1:
	vrcph	vtmp[0],    W12i[3]			/*Resv*/
	vrcpl	invW12f[3], W12f[3]			/*Resv*/
caseLXRej_G_VTX_Return2:				sbv	vfog12[15], oRSP_POINT_A-RSP_POINT_LEN*1+8(sys0)   /*FOG2*/
	#----------------------------------------------
	#  Z 値を 0 以上でクランプする
	#----------------------------------------------
#ifndef	NO_CLAMP_Z
  Assign(zscn12i, v28)
	vge	zscn12i, vscn12i, _0x0000		/*S*/
#endif
	vrcph	invW12i[3], W12i[7]			sbv	vfog12[7],  oRSP_POINT_A-RSP_POINT_LEN*2+8(sys0)   /*FOG1*/
	vrcpl	invW12f[7], W12f[7]			ssv	yscn12i[10], oRSP_POINT_YS-RSP_POINT_LEN*1(dest2)  /*Ys2*/
	vrcph	invW12i[7], _0x0000			ssv	yscn12i[2],  oRSP_POINT_YS-RSP_POINT_LEN*2(dest2)  /*Ys1*/
  EndAssign(yscn12i, v2)
  EndAssign(vfog12, v27)

#else		/* EXRej */
	
caseLXRej_G_VTX_Return:
	#-------------------------------------------------------------
	#  w に PerspNorm(=n) の値を乗する.
	#	これにより 1/w の値を求めるときの	
	#	Overflow を防止する.
	#  前ループのスクリーン座標値の出力
	#-------------------------------------------------------------
  Assign(cflag, 2)
  Assign(W12f, v3)
  Assign(W12i, v4)
	vmudl	vtmp, vout12f, pNorm			/*S*/
	vmadm	W12i, vout12i, pNorm			/*S*/
	vmadn	W12f, vzero,   _0x0000			addi	num, num, -2*2		# num 値を減算する
	#----------------------------------------------
	#  クリップ用パラメータ作成	
	#----------------------------------------------
  Assign(vclip12f, v5)
  Assign(vclip12i, v6)
	vmudn	vclip12f, vout12f, clipBox		sub	sys0, dest2, fogon	# Fog=OFF なら Ys の位置に書き込む.
	vmadh	vclip12i, vout12i, clipBox		sbv	vfog12[15], oRSP_POINT_A-RSP_POINT_LEN*1+8(sys0)    /*FOG2*/
	#----------------------------------------------
	#  Z 値を 0 以上でクランプする
	#----------------------------------------------
#ifndef	NO_CLAMP_Z
  Assign(zscn12i, v28)
	vge	zscn12i, vscn12i, _0x0000		/*S*/
#endif
	#----------------------------------------------
	#  1/nw の値を Newton 法で求める (1)
	#	vrcph/l で求めるのは 1/2X の値
	#	(R=1/X)
	#	1/X = R*(2-R*X) = R/2*(4+R/2*X*(-4))
	#----------------------------------------------
  Assign(invW12f, v7)
  Assign(invW12i, v21)
	vrcph	vtmp[0],    W12i[3]			sbv	vfog12[7],  oRSP_POINT_A-RSP_POINT_LEN*2+8(sys0)    /*FOG1*/
#ifndef	NO_CLAMP_Z
	vrcpl	invW12f[3], W12f[3]			slv	vscn12i[8], oRSP_POINT_XS-RSP_POINT_LEN*1(dest2)    /*XYs2*/
	vrcph	invW12i[3], W12i[7]			slv	vscn12i[0], oRSP_POINT_XS-RSP_POINT_LEN*2(dest2)    /*XYs1*/
#else
	vrcpl	invW12f[3], W12f[3]			sdv	vscn12i[8], oRSP_POINT_XS-RSP_POINT_LEN*1-4(dest2x) /*XYZs2*/
	vrcph	invW12i[3], W12i[7]			sdv	vscn12i[0], oRSP_POINT_XS-RSP_POINT_LEN*2(dest2)    /*XYZs1*/
#endif
	vrcpl	invW12f[7], W12f[7]			ssv	vscn12f[12],oRSP_POINT_ZSF-RSP_POINT_LEN*1(dest2)   /*Zsf2*/
	vrcph	invW12i[7], _0x0000			ssv	vscn12f[4], oRSP_POINT_ZSF-RSP_POINT_LEN*2(dest2)   /*Zsf1*/
  EndAssign(vfog12, v27)
#endif
	#----------------------------------------------
	#  クリップ判定	(1)
	#	n*W と X,Y,Z,-X,-Y,-Z を比較
	#----------------------------------------------
#ifndef	NO_CLAMP_Z
	vch	vtmp,  vout12i, vclip12i[3h]		ssv	zscn12i[12],oRSP_POINT_ZS-RSP_POINT_LEN*1(dest2)    /*Zs2*/
	vcl	vtmp,  vout12f, vclip12f[3h]		ssv	zscn12i[4], oRSP_POINT_ZS-RSP_POINT_LEN*2(dest2)    /*Zs1*/
  EndAssign(zscn12i, v28)
#else
	vch	vtmp,  vout12i, vclip12i[3h]		/*S*/
	vcl	vtmp,  vout12f, vclip12f[3h]		/*S*/
#endif
  EndAssign(vclip12f, v5)
  EndAssign(vclip12i, v6)
	#----------------------------------------------
	#  1/nw の値を Newton 法で求める (2)
	#----------------------------------------------
  Assign(vcolor, v2)
	vmudl	vtmp, W12f, invW12f			ldv	vcolor[0],   8(src)		# カラー値の取得
	vmadm	vtmp, W12i, invW12f			/*-*/	cfc2	cflag, $vcc		# Reject 判定値取得
	vmadn	W12f, W12f, invW12i			sra	sys0,  num,  31			# dest+=20,dest2=dest とする.
	vmadh	W12i, W12i, invW12i			andi	sys0, sys0, RSP_POINT_LEN	# もしnum<0 なら出力を重ねる.
	#----------------------------------------------
	#  Z 値の W へのコピー (1)
	#	$vcc = 01110111 となり [3|7]=0
	#----------------------------------------------
	vlt	vtmp, vconst0, _0x7f00			addi	dest, dest, RSP_POINT_LEN*2
	#----------------------------------------------
	#  1/nw の値を Newton 法で求める (3)
	#----------------------------------------------
  Assign(vrej12i, v5)
	vmudh	vtmp, vone, _0x0004			sub	dest2, dest, sys0
	vmadn	W12f, W12f, _0xfffc			/*-*/	andi	sys0, cflag, 0x0707
	vmadh	W12i, W12i, _0xfffc			/*-*/	andi	sys1, cflag, 0x7070
  EndAssign(cflag, 2)
	#----------------------------------------------
	#  Z 値の W へのコピー (2)
	#----------------------------------------------
	vmrg	vout12f, vout12f, vout12f[2h]		/*-*/	mtc2	sys0, vrej12i[0]
	vmrg	vout12i, vout12i, vout12i[2h]		/*-*/	mtc2	sys1, vrej12i[8]
	#----------------------------------------------
	#  1/nw の値を Newton 法で求める (4)
	#----------------------------------------------	
	vmudl	vtmp,     W12f, invW12f			/*S*/
	vmadm	vtmp,     W12i, invW12f			addi	src, src, 32			# src ポインタ加算
	vmadn	invW12f,  W12f, invW12i			slv	vcolor[0], oRSP_POINT_R-RSP_POINT_LEN*1(dest2)	/*Clr2*/
	vmadh	invW12i,  W12i, invW12i			slv	vcolor[4], oRSP_POINT_R-RSP_POINT_LEN*2(dest)	/*Clr1*/
  EndAssign(vcolor, v2)
  EndAssign(W12f, v3)
  EndAssign(W12i, v4)
	#----------------------------------------------
	#  テクスチャ座標値のスケーリング
	#----------------------------------------------
	vmudm	st12, st12, txscale			/*S*/
	/*V*/						/*S*/
	#----------------------------------------------
	#  透視変換を行なう (x,y,z を w で割る)
	#	1/nW を (x,y,z) に掛ける
	#----------------------------------------------
	vmudl	vtmp,    vout12f, invW12f[3h]		/*S*/
	vmadm	vtmp,    vout12i, invW12f[3h]		/*S*/
	vmadn	vout12f, vout12f, invW12i[3h]		/*S*/
	vmadh	vout12i, vout12i, invW12i[3h]		ssv	invW12f[14],oRSP_POINT_INVWF-RSP_POINT_LEN*1(dest2)/*invWf2*/
	#----------------------------------------------
	#  Reject 判定値を計算する
	#	vrej12i が正なら 0x7fff
	#----------------------------------------------
	veq	vtmp,    vzero, vrej12i[0h]		ssv	invW12f[6], oRSP_POINT_INVWF-RSP_POINT_LEN*2(dest) /*invWf1*/
	vmrg	vrej12i, vzero, _0x7fff			ssv	invW12i[14],oRSP_POINT_INVWI-RSP_POINT_LEN*1(dest2)/*invWi2*/
	#----------------------------------------------
	#  PerspNorm を掛け, (x/w,y/w,z/w) を求める
	#----------------------------------------------
	vmudl	vtmp,    vout12f, pNorm			ssv	invW12i[6],oRSP_POINT_INVWI-RSP_POINT_LEN*2(dest)/*invWf1*/
	vmadm	vout12i, vout12i, pNorm			slv	st12[4],   oRSP_POINT_S-RSP_POINT_LEN*1(dest2) /*ST2*/
	vmadn	vout12f, vzero,   _0x0000		slv	st12[12],  oRSP_POINT_S-RSP_POINT_LEN*2(dest)  /*ST1*/
  EndAssign(st12, v22)
  EndAssign(invW12f, v7)
  EndAssign(invW12i, v21)
	#----------------------------------------------
	#  スクリーン座標系に変換する
	#	fog 値の計算も同時に行なう.
	#	このときの fog 計算値は 0x7f00 を
	#	加算している.
	#----------------------------------------------
	vmudh	vtmp,    vrej12i, _0x7fff		ldv	vin12i[0],  0(src)	# 次の頂点データ1ロード
	vmadh	vtmp,    vptrans, _0x0001		ldv	vin12i[8], 16(src)	# 次の頂点データ2ロード
	vmadh	vtmp,    vpfog,   _0x7f00		addi	dest2x, dest2, 4
	vmadn	vscn12f, vout12f, vpscale		bgtz	num, caseLXRej_G_VTX_Loop
	vmadh	vscn12i, vout12i, vpscale 		/*Delay*/
  EndAssign(vrej12i, v5)
	#----------------------------------------------
	#  フォグ値のクランプ
	#  Y 座標のサブピクセルのマスク
	#----------------------------------------------
  Assign(vfog12,  v2)
  Assign(yscn12i, v3)
  Assign(zscn12i, v28)
#ifdef	NOSUBPIX	/* LXRej */
	/*V*/						sub	sys1, dest2, fogon
	/*V*/						sub	sys0, dest,  fogon
	/*V*/						/*S*/
# ifndef NO_CLAMP_Z
	vge	zscn12i, vscn12i, _0x0000		ssv	vscn12i[8],  oRSP_POINT_XS-RSP_POINT_LEN*1(dest2)  /*Xs2*/
	vge	vfog12,  vscn12i, _0x7f00		ssv	vscn12i[0],  oRSP_POINT_XS-RSP_POINT_LEN*2(dest)   /*Xs1*/
	vand	yscn12i, vscn12i, _0xfffc		ssv	vscn12f[12], oRSP_POINT_ZSF-RSP_POINT_LEN*1(dest2) /*Zsf2*/
	/*V*/						ssv	vscn12f[4],  oRSP_POINT_ZSF-RSP_POINT_LEN*2(dest)  /*Zsf1*/
	/*V*/						ssv	zscn12i[12], oRSP_POINT_ZS-RSP_POINT_LEN*1(dest2)  /*Zs2*/
	/*V*/						ssv	zscn12i[4],  oRSP_POINT_ZS-RSP_POINT_LEN*2(dest)   /*Zs1*/							
# else
	vge	vfog12,  vscn12i, _0x7f00		sdv	vscn12i[8],  oRSP_POINT_XS-RSP_POINT_LEN*1-4(dest2x)/*XZs2*/
	vand	yscn12i, vscn12i, _0xfffc		sdv	vscn12i[0],  oRSP_POINT_XS-RSP_POINT_LEN*2(dest)   /*XZs1*/
	/*V*/						ssv	vscn12f[12], oRSP_POINT_ZSF-RSP_POINT_LEN*1(dest2) /*Zsf2*/
	/*V*/						ssv	vscn12f[4],  oRSP_POINT_ZSF-RSP_POINT_LEN*2(dest)  /*Zsf1*/
# endif
	/*V*/						sbv	vfog12[15],  oRSP_POINT_A-RSP_POINT_LEN*1+8(sys1)  /*FOG2*/
	/*V*/						sbv	vfog12[7],   oRSP_POINT_A-RSP_POINT_LEN*2+8(sys0)  /*FOG1*/
	/*V*/						ssv	yscn12i[10], oRSP_POINT_YS-RSP_POINT_LEN*1(dest2)  /*Ys2*/
	/*V*/						j	GfxDone
	/*Delay*/					ssv	yscn12i[2],  oRSP_POINT_YS-RSP_POINT_LEN*2(dest)   /*Ys1*/
#else			/* EXRej */
	/*V*/						/*S*/
	/*V*/						/*S*/
	/*V*/						/*S*/
# ifndef NO_CLAMP_Z
	vge	zscn12i, vscn12i, _0x0000		slv	vscn12i[8],  oRSP_POINT_XS -RSP_POINT_LEN*1(dest2) /*Xs2*/
	vge	vfog12,  vscn12i, _0x7f00		slv	vscn12i[0],  oRSP_POINT_XS -RSP_POINT_LEN*2(dest)  /*Xs1*/
	/*V*/						ssv	vscn12f[12], oRSP_POINT_ZSF-RSP_POINT_LEN*1(dest2) /*Zsf2*/
	/*V*/						ssv	vscn12f[4],  oRSP_POINT_ZSF-RSP_POINT_LEN*2(dest)  /*Zsf1*/
	/*V*/						ssv	zscn12i[12], oRSP_POINT_ZS -RSP_POINT_LEN*1(dest2) /*Zs2*/
	/*V*/						beq	fogon, zero, GfxDone
	/*Delay*/					ssv	zscn12i[4],  oRSP_POINT_ZS -RSP_POINT_LEN*2(dest)  /*Zs1*/
	/*V*/						sbv	vfog12[15],  oRSP_POINT_A  -RSP_POINT_LEN*1(dest2) /*FOG2*/
	/*V*/						j	GfxDone
	/*Delay*/					sbv	vfog12[7],   oRSP_POINT_A  -RSP_POINT_LEN*2(dest)  /*FOG1*/
# else
	vge	vfog12,  vscn12i, _0x7f00		sdv	vscn12i[8],  oRSP_POINT_XS-RSP_POINT_LEN*1-4(dest2x)/*XZs2*/
	/*V*/						sdv	vscn12i[0],  oRSP_POINT_XS-RSP_POINT_LEN*2(dest)   /*XZs1*/
	/*V*/						ssv	vscn12f[12], oRSP_POINT_ZSF-RSP_POINT_LEN*1(dest2) /*Zsf2*/
	/*V*/						beq	fogon, zero, GfxDone
	/*Delay*/					ssv	vscn12f[4],  oRSP_POINT_ZSF-RSP_POINT_LEN*2(dest)  /*Zsf1*/
	/*V*/						sbv	vfog12[15],  oRSP_POINT_A-RSP_POINT_LEN*1(dest2)   /*FOG2*/
	/*V*/						j	GfxDone
	/*Delay*/					sbv	vfog12[7],   oRSP_POINT_A-RSP_POINT_LEN*2(dest)    /*FOG1*/
#  endif
#endif	
  EndAssign(yscn12i, v3)
  EndAssign(zscn12i, v28)
  EndAssign(num, 1)
  EndAssign(src, 14)
  EndAssign(dest, 15)
  EndAssign(gmode_h, 5)
  EndAssign(lighton, 6)
  EndAssign(fogon, 7)
  EndAssign(dest2, 8)
  EndAssign(dest2x, 10)
  EndAssign(lstat, 13)
  EndAssign(mtx0, v8)
  EndAssign(mtx1, v9)
  EndAssign(mtx2, v10)
  EndAssign(mtx3, v11)
  EndAssign(mtf0, v12)
  EndAssign(mtf1, v13)
  EndAssign(mtf2, v14)
  EndAssign(mtf3, v15)
  EndAssign(vpscale, v16)
  EndAssign(vptrans, v17)
  EndAssign(txscale, v18)
  EndAssign(vpfog, v19)
  EndAssign(vin12i, v20)
  EndAssign(vout12f, v23)
  EndAssign(vout12i, v24)
  EndAssign(vscn12f, v26)
  EndAssign(vscn12i, v25)
  EndAssign(vfog12, v2)

/*======== End of gxvtx_lxrej.s ========*/
